#ifndef RTOS_TYPES_H
#define RTOS_TYPES_H
/*
* Copyright (c) Andras Zsoter 2014-2015.
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in
* all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
* THE SOFTWARE.
*
*/

#include <stdint.h>

// Type to store saved flags e.g. when we enter and exit critical state.
typedef uint32_t RTOS_Critical_State;

typedef uint32_t RTOS_StackItem_t;

struct rtos_StackFrame
{
	RTOS_StackItem_t spsr;
	RTOS_StackItem_t regs[16];
};

typedef struct rtos_StackFrame rtos_StackFrame;

// This is needed when we create stack frames for tasks that haven't started yet.
#define RTOS_INITIAL_STACK_DEPTH (sizeof(rtos_StackFrame) + 8)


#if ((RTOS_Priority_Highest) > 31)
#	define RTOS_TASKSET_TYPE uint64_t
#	define RTOS_HIGHEST_SUPPORTED_TASK_PRIORITY 63
#	define RTOS_FIND_HIGHEST(X) (63 - __builtin_clzll((uint64_t)(X)))
#		define RTOS_TaskSet_NumberOfMembers(X) __builtin_popcountll((uint64_t)(X))
#else
#	if !defined(RTOS_FIND_HIGHEST)
#		define RTOS_FIND_HIGHEST(X) (31 - rtos_CLZ(X))
#		define RTOS_TaskSet_NumberOfMembers(X) __builtin_popcountl(X)
#	endif
#endif

#define RTOS_MIN_STACK_SIZE 512 /* What is the real number? */

#if defined(RTOS_SMP)
typedef uint32_t RTOS_CpuId;
#define RTOS_CPUID_NO_CPU ((RTOS_CpuId)(-1))
typedef uint32_t RTOS_CpuMutex;
typedef uint32_t RTOS_CpuMask;
#endif

#endif

