#include "xparameters.h"
// #include "xil_cache.h"

#include "platform_config.h"
// #include "xparameters.h"
#include "xscutimer.h"
#include "xscugic.h"

#include "rtos.h"

XScuGic InterruptController = { IsReady: 1 };
// static XScuGic InterruptController; 	/* Interrupt controller instance */
static XScuTimer Timer;						/* A9 timer counter */

#define XSCUTIMER_CLOCK_HZ XPAR_CPU_CORTEXA9_0_CPU_CLK_FREQ_HZ/2
#define TICKS_PER_SECOND RTOS_TICKS_PER_SECOND

static void SetupInterruptController(void)
{
	int Status;
	XScuGic_Config *IntcConfig; /* The configuration parameters of the
									interrupt controller */
	/*
	 * Initialize the interrupt controller driver
	 */
	IntcConfig = XScuGic_LookupConfig(XPAR_SCUGIC_SINGLE_DEVICE_ID);
	if (IntcConfig == NULL) {
		return;
	}

	Status = XScuGic_CfgInitialize(&InterruptController, IntcConfig,
									IntcConfig->CpuBaseAddress);
	if (Status != XST_SUCCESS) {
		return;
	}

	Xil_ExceptionRegisterHandler(XIL_EXCEPTION_ID_IRQ_INT,
						(Xil_ExceptionHandler)XScuGic_InterruptHandler,
						&InterruptController);

}

extern void rtos_TimerTick(void);

static void TimerHandler(XScuTimer *Timer)
{
	rtos_TimerTick();

	XScuTimer_ClearInterruptStatus(Timer);
}


/*
 * Setup the A9 internal timer to generate the tick interrupts at the
 * required frequency.
 */
static void SetupTimerInterrupt(void)
{
	int Status;
	XScuTimer_Config *ScuConfig;

	// SetupInterruptController();
#if 0
	Xil_ExceptionRegisterHandler(XIL_EXCEPTION_ID_IRQ_INT,
						(Xil_ExceptionHandler)XScuGic_InterruptHandler,
						&InterruptController);
#endif

	/*
	 * Connect to the interrupt controller
	 */
	Status = XScuGic_Connect(&InterruptController, XPAR_SCUTIMER_INTR,
						(Xil_ExceptionHandler)TimerHandler, (void *)&Timer);
	if (Status != XST_SUCCESS) {
		return;
	}

	/* Timer Setup */

	/*
	 * Initialize the A9Timer driver.
	 */
	ScuConfig = XScuTimer_LookupConfig(XPAR_SCUTIMER_DEVICE_ID);

	Status = XScuTimer_CfgInitialize(&Timer, ScuConfig,
									ScuConfig->BaseAddr);
	if (Status != XST_SUCCESS) {
		return;
	}

	/*
	 * Enable Auto reload mode.
	 */
	XScuTimer_EnableAutoReload(&Timer);

	/*
	 * Load the timer counter register.
	 */
	XScuTimer_LoadTimer(&Timer, XSCUTIMER_CLOCK_HZ / (TICKS_PER_SECOND));

	/*
	 * Start the timer counter and then wait for it
	 * to timeout a number of times.
	 */
	XScuTimer_Start(&Timer);

	/*
	 * Enable the interrupt for the Timer in the interrupt controller
	 */
	XScuGic_Enable(&InterruptController, XPAR_SCUTIMER_INTR);

	/*
	 * Enable the timer interrupts for timer mode.
	 */
	XScuTimer_EnableInterrupt(&Timer);

	/*
	 * Do NOT enable interrupts in the ARM processor here.
	 * This happens when the scheduler is started.
	 */
}

#if defined(RTOS_SMP)
#define RTOS_SGI_INT_ID 0
// Software Generated Interrupt (SGI)
static void SgiHandler(void *dummy)
{
	(void)dummy;
}

#if 0
void SetupGIC(void)
{
	XScuGic_CPUWriteReg(&InterruptController, XSCUGIC_CPU_PRIOR_OFFSET, 0xF0);
	XScuGic_CPUWriteReg(&InterruptController, XSCUGIC_CONTROL_OFFSET, 0x07);
}
#endif

void EnableSgi(void)
{
	XScuGic_Enable(&InterruptController, RTOS_SGI_INT_ID);
}

void SetupSoftwareInterrupt(void)
{
	int Status;
	Status = XScuGic_Connect(&InterruptController, RTOS_SGI_INT_ID, (Xil_ExceptionHandler)SgiHandler, (void *)0);

	if (Status != XST_SUCCESS)
	{
		return;
	}

	EnableSgi();
}

void Board_SendSoftwareInterrupt(uint32_t cpus)
{
	XScuGic_SoftwareIntr(&InterruptController, RTOS_SGI_INT_ID, cpus);
}
#endif

void bsp_init(void)
{
	SetupInterruptController();
#if defined(RTOS_SMP)
	SetupSoftwareInterrupt();
#endif
	SetupTimerInterrupt();
}

void Board_HardwareInit(void)
{
	bsp_init();
}

void board_HandleIRQ(void)
{
	__asm volatile( "bl IRQInterrupt" ); 
}


