/*
* Copyright (c) 2015 Andras Zsoter
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in
* all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
* THE SOFTWARE.
*
*/

#include "forth.h"
#include "forth_interface.h"
#include <rtos.h>

#if defined(FORTH_INCLUDE_MS)
void forth_ms(struct forth_runtime_context *rctx)
{
	forth_cell_t dly = FORTH_POP(rctx);
	// xil_printf("MS: dly=%d which is %d ticks.\r\n", dly, RTOS_MS_TO_TICKS(dly));
	RTOS_Delay(RTOS_MS_TO_TICKS(dly));
}
#endif

#if defined(FORTH_EXTERNAL_PRIMITIVES)
#if defined(RTOS_SMP)
forth_cell_t forth_current_cpu(forth_runtime_context_p rctx)
{
	FORTH_PUSH(rctx, RTOS_CurrentCpu());
	return 0;
}
#endif

forth_cell_t forth_dot_os(forth_runtime_context_p rctx)
{
	rtos_Debug();
	return 0;
}

forth_external_primitive external_primitive_table[] =
{
		forth_dot_os,
#if defined(RTOS_SMP)
		forth_current_cpu,
#endif
};


static void register_primitive(struct forth_runtime_context *rctx, const char *name, forth_cell_t index)
{
	xil_printf("Registering %s at %d.\r\n", name, index);

	int res = forth_register_external_primitive(rctx, name, index);

	if (0 != res)
	{
		xil_printf("Return value = %d at item %d (%s).\n", res, index, name);
	}
}

static int already_registered = 0;

void printRegistered(void)
{
	xil_printf("already_registered = 0x%x.\r\n", already_registered);
}

int forth_setup_application_primitives(struct forth_runtime_context *rctx)
{
	int i;
		rctx->external_primitive_table = external_primitive_table;

		xil_printf("\r\nFORTH: Registering external primitives.\r\n");

		printRegistered();

		if (already_registered)
		{
			return 0;
		}

		print("##1\r\n");
		already_registered = 1;
		print("##2\r\n");
		for (i = 0; i < sizeof(external_primitive_table) / sizeof(forth_external_primitive); i++)
		{
			xil_printf("i=%d\r\n", i);

			if (forth_dot_os == external_primitive_table[i])
			{
				register_primitive(rctx, ".OS", i);
			}
#if defined(RTOS_SMP)
			else if (forth_current_cpu == external_primitive_table[i])
			{
				register_primitive(rctx, "CPU@", i);
			}
#endif
			else
			{
				printf("Undefined external primitive at index = %d.\n", i);
			}
		}

		return 0;
}
#endif
