/*
* Copyright (c) Andras Zsoter 2014.
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in
* all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
* THE SOFTWARE.
*
*/

#include <stdio.h>
#include "platform.h"
#include <stdint.h>
#include "rtos.h"

#include "xparameters.h"
#include "lwipopts.h"

#include "platform_config.h"
#include "platform.h"

#include "lwip/sockets.h"
#include "lwip/sys.h"
#include "lwip/init.h"
#include "netif/xadapter.h"
#include "lwip/dhcp.h"

#if __arm__
#include "xil_printf.h"
#endif

#include "board.h"

#define RTOS_Priority_MainTask 2

extern int forth_run_serial();

#define MS_TO_TICKS(MS) RTOS_MS_TO_TICKS((MS))


#define THREAD_STACKSIZE ((TCPIP_THREAD_STACKSIZE) / sizeof(RTOS_StackItem_t))


RTOS_StackItem_t mainTaskStack[THREAD_STACKSIZE];
RTOS_StackItem_t stackIdle[RTOS_MIN_STACK_SIZE];
RTOS_StackItem_t stackExp[RTOS_MIN_STACK_SIZE] = { 1 };

RTOS_Task mainTask;
RTOS_Task taskIdle;
RTOS_Task taskExp;

extern int socket_error(int sd);

void print_ip(char *msg, struct ip_addr *ip)
{
    print(msg);
    xil_printf("%d.%d.%d.%d\r\n", ip4_addr1(ip), ip4_addr2(ip),
            ip4_addr3(ip), ip4_addr4(ip));
}

void print_ip_settings(struct ip_addr *ip, struct ip_addr *mask, struct ip_addr *gw)
{

    print_ip("Board IP: ", ip);
    print_ip("Netmask : ", mask);
    print_ip("Gateway : ", gw);
}

//------------------------------------------------------------------------------------
u16_t telnet_port = 23;

void print_telnet_app_header()
{
    xil_printf("%20s %6d\r\n", "Telnet server", telnet_port);

}

void print_headers()
{
	print_telnet_app_header();
}

char req[] = { 255, 254, 1, 255, 251, 1 }; // DON'T ECHO, WON'T ECHO

extern int forth_run_forth_network(int sd, unsigned int dstack_cells, unsigned int rstack_cells);

#if defined(RTOS_SMP)
void print_cpu(void)
{
	outbyte('[');
	outbyte('0' + RTOS_CurrentCpu());
	outbyte(']');
}
#endif

void handle_telnet_request(void *p)
{
	int non_blocking = 1;
	int sd = (int)p;
#if defined(RTOS_SMP)
	print_cpu();
#endif
	xil_printf("sd = %d\r\n", sd);

	write(sd, req, sizeof(req));
	lwip_ioctl(sd, FIONBIO, &non_blocking);
	lwip_fcntl(sd, F_SETFL, lwip_fcntl(sd, F_GETFL, 0) | O_NONBLOCK);
	forth_run_forth_network(sd, 512, 512);

	xil_printf("Closing socket: %d\r\n", sd);
	close(sd);
}

void telnet_application_thread(void *p)
{
	int sock, new_sd;
	struct sockaddr_in address, remote;
	int size;

#if defined(RTOS_SMP)
			print_cpu();
#endif

	xil_printf("Telnet...\r\n");

	if ((sock = lwip_socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
		xil_printf("Failed to create server socket. Bailing out.....\r\n");
		return;
	}

	address.sin_family = AF_INET;
	address.sin_port = htons(telnet_port);
	address.sin_addr.s_addr = INADDR_ANY;

	if (lwip_bind(sock, (struct sockaddr *)&address, sizeof (address)) < 0)
	{
		xil_printf("Failed to bind to server socket. Bailing out.....\r\n");
		close(sock);
		return;
	}

	lwip_listen(sock, 0);

	size = sizeof(remote);

	while (1)
	{
		if ((new_sd = lwip_accept(sock, (struct sockaddr *)&remote, (socklen_t *)&size)) > 0)
		{
#if defined(RTOS_SMP)
			print_cpu();
#endif
			xil_printf("New client sd= %d\r\n", new_sd);
			sys_thread_new("telnet", handle_telnet_request,	(void*)new_sd, TCPIP_THREAD_STACKSIZE, DEFAULT_THREAD_PRIO);
		}
		else
		{
			xil_printf("error = %d\r\n",socket_error(sock));
		}
	}
	(void)p;
}

void forth_serial_thread(void *p)
{
	(void)p;
	while (1)
	{
		forth_run_serial();
	}
}

void test(void *p)
{
	while(1)
	{
		//outbyte('+');
		// RTOS_YieldTimeSlice();
		// RTOS_Delay(2);
	}
	(void)p;
}

void launch_app_threads()
{
    // sys_thread_new("test", test, 0, TCPIP_THREAD_STACKSIZE, DEFAULT_THREAD_PRIO);
    sys_thread_new("serial-console", forth_serial_thread, 0, TCPIP_THREAD_STACKSIZE, DEFAULT_THREAD_PRIO);
    xil_printf("Starting telnet server....\r\n");
    sys_thread_new("telnet-server", telnet_application_thread, 0, TCPIP_THREAD_STACKSIZE, DEFAULT_THREAD_PRIO);
}

//------------------------------------------------------------------------------------

struct netif server_netif;

uint32_t getMyIp(void)
{
	return ntohl(server_netif.ip_addr.addr);
}


void network_thread(void *p)
{
    struct netif *netif;
    struct ip_addr ipaddr, netmask, gw;
#if LWIP_DHCP==1
    int mscnt = 0;
#endif
    /* the mac address of the board. this should be unique per board */
    unsigned char mac_ethernet_address[] = { 0x00, 0x0a, 0x35, 0x00, 0x01, 0x02 };
    // RTOS_SavedCriticalState(saved_state);

    netif = &server_netif;
    xil_printf("Network Thread.\r\n");

#if LWIP_DHCP==0
    /* initliaze IP addresses to be used */
    IP4_ADDR(&ipaddr,  192, 168,   1, 10);
    IP4_ADDR(&netmask, 255, 255, 255,  0);
    IP4_ADDR(&gw,      192, 168,   1,  1);
#endif

    /* print out IP settings of the board */
    print("\r\n\r\n");
    print("-----lwIP Socket Mode Demo Application ------\r\n");

#if LWIP_DHCP==0
    print_ip_settings(&ipaddr, &netmask, &gw);
    /* print all application headers */
#endif

#if LWIP_DHCP==1
	ipaddr.addr = 0;
	gw.addr = 0;
	netmask.addr = 0;
#endif
    /* Add network interface to the netif_list, and set it as default */
    if (!xemac_add(netif, &ipaddr, &netmask, &gw, mac_ethernet_address, PLATFORM_EMAC_BASEADDR))
    {
        xil_printf("Error adding N/W interface\r\n");
        return;
    }
    netif_set_default(netif);

    /* specify that the network if is up */
    netif_set_up(netif);

    /* start packet receive thread - required for lwIP operation */
    sys_thread_new("xemacif_input_thread", (void(*)(void*))xemacif_input_thread, netif,	TCPIP_THREAD_STACKSIZE, DEFAULT_THREAD_PRIO);

#if LWIP_DHCP==1
    dhcp_start(netif);

    while (1)
    {
    	RTOS_Delay(MS_TO_TICKS(DHCP_FINE_TIMER_MSECS));
    	dhcp_fine_tmr();
    	mscnt += DHCP_FINE_TIMER_MSECS;
    	if (mscnt >= DHCP_COARSE_TIMER_SECS*1000)
    	{
			dhcp_coarse_tmr();
			mscnt = 0;
    	}
	}
#else
    print_headers();
    launch_app_threads();
#endif
    return;
}

char input_buffer[256];

void main_thread(void *p)
{
#if LWIP_DHCP==1
	int mscnt = 0;
#endif
#if defined(RTOS_SMP)
	// RTOS.TimeshareParallelAllowed = 1;
#endif
	xil_printf("Main Thread...\r\n");
	xil_printf("lwip_init()...\r\n");
	/* initialize lwIP before calling sys_thread_new */
    lwip_init();
    xil_printf("\r\nDone.\r\n");

    xil_printf("Running Network Thread...\r\n");

    /* any thread using lwIP should be created using sys_thread_new */
    sys_thread_new("NW_THRD", network_thread, NULL, THREAD_STACKSIZE, DEFAULT_THREAD_PRIO);

#if LWIP_DHCP==1
    while (1)
    {
    	RTOS_Delay(MS_TO_TICKS(DHCP_FINE_TIMER_MSECS));
    	outbyte('W');
		if (server_netif.ip_addr.addr)
		{
			xil_printf("DHCP request success\r\n");
			print_ip_settings(&(server_netif.ip_addr), &(server_netif.netmask), &(server_netif.gw));
			/* print all application headers */
			print_headers();
			/* now we can start application threads */
			launch_app_threads();
			break;
		}
		mscnt += DHCP_FINE_TIMER_MSECS;
		if (mscnt >= 10000)
		{
			xil_printf("ERROR: DHCP request timed out\r\n");
			xil_printf("Configuring default IP of 192.168.1.10\r\n");
			IP4_ADDR(&(server_netif.ip_addr),  192, 168,   1, 10);
			IP4_ADDR(&(server_netif.netmask), 255, 255, 255,  0);
			IP4_ADDR(&(server_netif.gw),      192, 168,   1,  1);
			print_ip_settings(&(server_netif.ip_addr), &(server_netif.netmask), &(server_netif.gw));
			/* print all application headers */
			print_headers();
			launch_app_threads();
			break;
		}

	}

#endif
    (void)p;
    return;
}

extern void rtos_debug_PrintOS(void);

void Exp(void *p)
{
	while(1)
	{
#if defined(RTOS_SMP)
//		rtos_debug_PrintOS();
//		print_cpu();
#endif
		__asm__ __volatile__ ("WFI");
	}
	(void)p;
}

extern int  _heap_start;
extern int _heap_end;

extern u8 _end;
#include <lwip/mem.h>
#define MEM_SIZE_ALIGNED     LWIP_MEM_ALIGN_SIZE(MEM_SIZE)
#define SIZEOF_STRUCT_MEM 64
extern u8_t ram_heap[MEM_SIZE_ALIGNED + (2*SIZEOF_STRUCT_MEM) + MEM_ALIGNMENT];

// extern volatile RTOS_OS RTOS __attribute__((nocommon, section(data)));
extern char __bss_start[];
extern char __bss_end[];
int main(void)
{
	char *p;
	char *endAdd = &_end;
	char *endAdd_aligned = (char *)(((int)endAdd + 0x100000) & (~0xFFFFF));
	int res;

	Board_HardwareInit();

	printRegistered();

#if 0
	Xil_DCacheFlush();
	// Xil_SetTlbAttributes((int)endAdd_aligned, 0xc02); // addr, attr
	// Xil_SetTlbAttributes((int)endAdd_aligned, 0x14de2);
	Xil_SetTlbAttributes((int)endAdd_aligned, 0x11de2);
	// Xil_SetTlbAttributes((int)(ram_heap), 0xc02); // addr, attr
	// Xil_SetTlbAttributes((int)(&_heap_start), 0xc02); // addr, attr
	// Xil_SetTlbAttributes((int)(&_heap_start),0x14de2);           // S=b1 TEX=b100 AP=b11, Domain=b1111, C=b0, B=b0

	for (p = __bss_start; p < __bss_end; p += 1048576)
	{
		xil_printf("p=0x%x\r\n",p);
		// Xil_SetTlbAttributes((int)p, 0xc02); // addr, attr
		// Xil_SetTlbAttributes((int)p,0x14de2);           // S=b1 TEX=b100 AP=b11, Domain=b1111, C=b0, B=b0
		Xil_SetTlbAttributes((int)p,0x11de2);
		// Xil_SetTlbAttributes((int)p,0x4de2);           // S=b1 TEX=b100 AP=b11, Domain=b1111, C=b0, B=b0
	}
#endif
	xil_printf("[JaeOS Telnet Demo] Press ENTER to continue.....\r\n");
	inbyte();
	xil_printf("&ram_heap=0x%x sizeof(ram_heap)=0x%x\r\n", &ram_heap, sizeof(ram_heap));
  	xil_printf("Heap Start = &0x%x Heap End = &0x%x\r\n", &_heap_start, &_heap_end);
  	xil_printf("__bss_start=x%x __bss_end=x%x\r\n", __bss_start, __bss_end);
  	xil_printf("&RTOS=x%x\r\n",&RTOS);
	xil_printf("Telnet on lwIP demo for JaeOS compiled on %s %s.\r\n", __DATE__, __TIME__);
	RTOS_CreateTask(&taskIdle, RTOS_Priority_Idle,  &stackIdle, RTOS_MIN_STACK_SIZE, &RTOS_DefaultIdleFunction, 0);
	RTOS_SetTaskName(&taskIdle, "Idle");
	RTOS_CreateTask(&taskExp, 1,  &stackExp, RTOS_MIN_STACK_SIZE, &Exp, 0);
	RTOS_SetTaskName(&taskExp, "Exp");
	RTOS_CreateTask(&mainTask, RTOS_Priority_MainTask, &mainTaskStack, THREAD_STACKSIZE, &main_thread, 0);
	RTOS_SetTaskName(&mainTask, "main_task");
#if defined(RTOS_SMP)
	RTOS_RestrictTaskToCpus(RTOS.TaskList[RTOS_Priority_MainTask], (1UL << RTOS_CurrentCpu()));
	RTOS_RestrictTaskToCpus(&taskExp, 0x02);
#endif
	res = RTOS_StartMultitasking();
	xil_printf("Something has gone wrong or we would not be here. RTOS_StartMultitasking() has returned %d.\r\n", res);
	return 0;
}

