#include <stdint.h>
#include <rtos.h>
#include <rtos_queue.h>
#include "board.h"
#include "utility.h"

// RTOS Demo App.
// Copyright (C) Andras Zsoter, 2014. All Rights Reserved.

#define RTOS_Priority_Producer		1
#define RTOS_Priority_Consumer1		2
#define RTOS_Priority_Consumer2		3

RTOS_StackItem_t stack1[512];
RTOS_StackItem_t stack2[512];
RTOS_StackItem_t stack3[512];
RTOS_StackItem_t stackIdle[RTOS_MIN_STACK_SIZE];

RTOS_Task task1;
RTOS_Task task2;
RTOS_Task task3;
RTOS_Task task_idle;

RTOS_Semaphore s;

char c = '*';

RTOS_Queue q;

void Consumer(void *p)
{
    void *goodie;
    (void)p;
    while(1) 
    {
	if (RTOS_OK == RTOS_Dequeue(&q, &goodie, RTOS_TIMEOUT_FOREVER))
	{
		Board_Puts(RTOS_GetTaskName(RTOS_CURRENT_TASK()));
		Board_Putc(':');
		PrintHex((uint32_t)goodie);
		Board_Putc('\r');
		Board_Putc('\n');
		RTOS_YieldPriority();
	}  
    }
}

void Producer(void *p)
{
    uint32_t stuff = 1;
    (void)p;
    while(1) 
    {
	stuff++;
	Board_Puts(RTOS_GetTaskName(RTOS_CURRENT_TASK()));
	Board_Putc(':');
	PrintHex(stuff);
	Board_Putc('\r');
	Board_Putc('\n');
	RTOS_Enqueue(&q, (void *)(stuff), 0);
    }
    (void)p;
}

extern void HardwareInit(void);

void *buff[16];

int main()
{
	init_platform();
    	RTOS_CreateTask(&task1,     RTOS_Priority_Producer, stack1, 512, &Producer, 0);
	RTOS_SetTaskName(&task1, "Producer");
    	RTOS_CreateTask(&task2,     RTOS_Priority_Consumer1, stack2, 512, &Consumer, 0);
	RTOS_SetTaskName(&task2, "Consumer1");
    	RTOS_CreateTask(&task3,     RTOS_Priority_Consumer2, stack3, 512,  &Consumer, 0);
	RTOS_SetTaskName(&task3, "Consumer2");
    	RTOS_CreateTask(&task_idle, RTOS_Priority_Idle,  stackIdle, RTOS_MIN_STACK_SIZE, &RTOS_DefaultIdleFunction, 0);
	RTOS_SetTaskName(&task_idle, "Idle");
	RTOS_CreateQueue(&q, buff, 16);

	Board_HardwareInit();
	RTOS_StartMultitasking();
	Board_Puts("Something has gone seriously wrong!");
    	while(1);
    	return 0;
}

void app_AssertFailed(const char *condition, const char *file, const char *line, const char *function)
{
	Board_Puts("\r\nAssertion '");
	Board_Puts(condition);
	Board_Puts("' has failed. @");
	Board_Puts(file);
	Board_Putc(':');
	Board_Puts(line);
	Board_Puts(" in function: ");
	Board_Puts(function);
	Board_Puts("().\r\n");
	while(1);
}

